@@ -63,6 +63,8 @@
 
 namespace reflex {
 
+extern const unsigned short codepages[][256];
+
 /// Input character sequence class for unified access to sources of input text.
 /**
 Description
@@ -343,8 +345,7 @@ class Input {
       wstring_(NULL),
       file_(NULL),
       istream_(NULL),
-      size_(0),
-      handler_(NULL)
+      size_(0)
   {
     init();
   }
@@ -359,7 +360,8 @@ class Input {
       uidx_(input.uidx_),
       ulen_(input.ulen_),
       utfx_(input.utfx_),
-      page_(input.page_)
+      page_(input.page_),
+      handler_(input.handler_)
   {
     std::memcpy(utf8_, input.utf8_, sizeof(utf8_));
   }
@@ -503,6 +505,7 @@ class Input {
     ulen_ = input.ulen_;
     utfx_ = input.utfx_;
     page_ = input.page_;
+    handler_ = input.handler_;
     std::memcpy(utf8_, input.utf8_, sizeof(utf8_));
     return *this;
   }
@@ -752,6 +755,7 @@ class Input {
     ulen_ = 0;
     utfx_ = 0;
     page_ = NULL;
+    handler_ = NULL;
     if (file_ != NULL)
       file_init(enc);
   }
