@@ -275,6 +275,8 @@ static int convert_hex(const char *pattern, size_t len, size_t& pos, convert_fla
     while (++k < len && s < hex + sizeof(hex) - 1 && (c = pattern[k]) != '}')
       *s++ = c;
     *s = '\0';
+    if (k >= len || pattern[k] != '}')
+      throw regex_error(regex_error::mismatched_braces, pattern, pos + 1);
   }
   else if (c == 'x' || (c == 'u' && (flags & convert_flag::u4)))
   {
@@ -299,6 +301,33 @@ static int convert_hex(const char *pattern, size_t len, size_t& pos, convert_fla
   return -1;
 }
 
+static int convert_oct(const char *pattern, size_t len, size_t& pos)
+{
+  char oct[9];
+  oct[0] = '\0';
+  size_t k = pos;
+  int c = pattern[k++];
+  if (k < len && pattern[k] == '{')
+  {
+    char *s = oct;
+    while (++k < len && s < oct + sizeof(oct) - 1 && (c = pattern[k]) != '}')
+      *s++ = c;
+    *s = '\0';
+    if (k >= len || pattern[k] != '}')
+      throw regex_error(regex_error::mismatched_braces, pattern, pos + 1);
+  }
+  if (oct[0] != '\0')
+  {
+    char *r;
+    unsigned long n = std::strtoul(oct, &r, 8);
+    if (*r != '\0' || n > 0x10FFFF)
+      throw regex_error(regex_error::invalid_class, pattern, pos);
+    pos = k;
+    return static_cast<int>(n);
+  }
+  return -1;
+}
+
 ////////////////////////////////////////////////////////////////////////////////
 //                                                                            //
 //  Definition name expansion                                                 //
@@ -311,10 +340,9 @@ static const std::string& expand(const std::map<std::string,std::string> *macros
   size_t k = pos++;
   while (pos < len && (std::isalnum(pattern[pos]) || pattern[pos] == '_' || (pattern[pos] & 0x80) == 0x80))
     ++pos;
-  std::string name;
-  name.append(&pattern[k], pos - k);
-  if (pos >= len && (pattern[pos] != '\\' || pattern[pos + 1] != '}') && pattern[pos] != '}')
+  if (pos >= len || (pattern[pos] == '\\' ? pattern[pos + 1] != '}' : pattern[pos] != '}'))
     throw regex_error(regex_error::undefined_name, pattern, pos);
+  std::string name(&pattern[k], pos - k);
   std::map<std::string,std::string>::const_iterator i = macros->find(name);
   if (i == macros->end())
     throw regex_error(regex_error::undefined_name, pattern, k);
@@ -384,8 +412,12 @@ static std::string convert_ranges(const char *pattern, size_t pos, ORanges<int>&
 
 static void expand_list(const char *pattern, size_t len, size_t& loc, size_t& pos, convert_flag_type flags, const std::map<size_t,std::string>& mod, const char *signature, const char *par, const std::map<std::string,std::string> *macros, std::string& regex)
 {
+  bool no_newline = false;
   if (pos + 1 < len && pattern[pos] == '^')
+  {
+    no_newline = true;
     ++pos;
+  }
   while (pos + 1 < len)
   {
     int c = pattern[pos];
@@ -408,6 +440,8 @@ static void expand_list(const char *pattern, size_t len, size_t& loc, size_t& po
           k = j;
           while (k < len && pattern[k] != '}')
             ++k;
+          if (k >= len)
+            throw regex_error(regex_error::mismatched_braces, pattern, pos);
           name.append(pattern, j, k - j);
         }
         else
@@ -426,6 +460,18 @@ static void expand_list(const char *pattern, size_t len, size_t& loc, size_t& po
         }
         pos = k;
       }
+      else if (c == 's' && (flags & convert_flag::notnewline))
+      {
+        if (is_modified(mod, 'u'))
+          regex.append(&pattern[loc], pos - loc - 1).append("\\t\\x0b-\\r\\x85\\p{Z}");
+        else
+          regex.append(&pattern[loc], pos - loc - 1).append("\\h\\x0b-\\r\\x85\\xa0");
+        loc = pos + 1;
+      }
+      else if (c == 'n')
+      {
+        no_newline = false;
+      }
     }
     else if (c == '[' && (pattern[pos + 1] == ':' || pattern[pos + 1] == '.' || pattern[pos + 1] == '='))
     {
@@ -436,32 +482,41 @@ static void expand_list(const char *pattern, size_t len, size_t& loc, size_t& po
     else if (c == '|' && pattern[pos + 1] == '|' && pos + 3 < len && pattern[pos + 2] == '[')
     {
       // character class union [abc||[def]]
+      if (!supports_escape(signature, '['))
+        throw regex_error(regex_error::invalid_class, pattern, pos + 1);
       pos += 3;
       expand_list(pattern, len, loc, pos, flags, mod, signature, par, macros, regex);
     }
     else if (c == '&' && pattern[pos + 1] == '&' && pos + 3 < len && pattern[pos + 2] == '[')
     {
       // character class intersection [a-z&&[^aeiou]]
+      if (!supports_escape(signature, '['))
+        throw regex_error(regex_error::invalid_class, pattern, pos + 1);
       pos += 3;
       expand_list(pattern, len, loc, pos, flags, mod, signature, par, macros, regex);
     }
     else if (c == '-' && pattern[pos + 1] == '-' && pos + 3 < len && pattern[pos + 2] == '[')
     {
       // character class subtraction [a-z--[aeiou]]
+      if (!supports_escape(signature, '['))
+        throw regex_error(regex_error::invalid_class, pattern, pos + 1);
       pos += 3;
       expand_list(pattern, len, loc, pos, flags, mod, signature, par, macros, regex);
     }
     ++pos;
-    if (pos >= len)
-      break;
-    if (pattern[pos] == ']')
+    if (pos >= len || pattern[pos] == ']')
       break;
   }
   if (pos >= len || pattern[pos] != ']')
     throw regex_error(regex_error::mismatched_brackets, pattern, loc);
+  if (no_newline && (flags & convert_flag::notnewline))
+  {
+    regex.append(&pattern[loc], pos - loc).append("\\n");
+    loc = pos;
+  }
 }
 
-static void insert_escape_class(const char *pattern, size_t& pos, const std::map<size_t,std::string>& mod, ORanges<int>& ranges)
+static void insert_escape_class(const char *pattern, size_t pos, const std::map<size_t,std::string>& mod, ORanges<int>& ranges)
 {
   int c = pattern[pos];
   char name[2] = { static_cast<char>(lowercase(c)), '\0' };
@@ -552,7 +607,7 @@ static int insert_escape(const char *pattern, size_t len, size_t& pos, convert_f
   {
     size_t k = pos + 3 + (pattern[pos] == '0');
     int wc = c - '0';
-    while (++pos < k && (c = pattern[pos]) >= '0' && c <= '7')
+    while (++pos < k && pos < len && (c = pattern[pos]) >= '0' && c <= '7')
       wc = 8 * wc + c - '0';
     c = wc;
     --pos;
@@ -583,6 +638,8 @@ static int insert_escape(const char *pattern, size_t len, size_t& pos, convert_f
       k = j;
       while (k < len && pattern[k] != '}')
         ++k;
+      if (k >= len)
+        throw regex_error(regex_error::mismatched_braces, pattern, pos);
       name.assign(&pattern[j], k - j);
     }
     else
@@ -653,6 +710,7 @@ static int insert_escape(const char *pattern, size_t len, size_t& pos, convert_f
     // \s is the same as \p{Space} but without newline \n
     insert_escape_class(pattern, pos, mod, ranges);
     ranges.erase('\n');
+    return -1;
   }
   else if (std::isalpha(c))
   {
@@ -675,6 +733,8 @@ static void insert_posix_class(const char *pattern, size_t len, size_t& pos, ORa
   char *name = buf;
   while (pos + 1 < len && name < buf + sizeof(buf) - 1 && (pattern[pos] != ':' || pattern[pos + 1] != ']'))
     *name++ = pattern[pos++];
+  if (pos + 1 >= len)
+    throw regex_error(regex_error::invalid_class, pattern, pos);
   *name = '\0';
   name = buf + (*buf == '^');
   if (name[1] != '\0')
@@ -920,12 +980,13 @@ static void insert_list(const char *pattern, size_t len, size_t& pos, convert_fl
           throw regex_error(regex_error::invalid_class_range, pattern, pos);
         ranges.insert(pc, c);
         range = false;
+        pc = -2;
       }
       else
       {
         ranges.insert(c);
+        pc = c;
       }
-      pc = c;
     }
     ++pos;
     if (pos >= len)
@@ -952,7 +1013,7 @@ static void insert_list(const char *pattern, size_t len, size_t& pos, convert_fl
 //                                                                            //
 ////////////////////////////////////////////////////////////////////////////////
 
-static void convert_escape_char(const char *pattern, size_t& loc, size_t& pos, convert_flag_type flags, const char *signature, const std::map<size_t,std::string>& mod, const char *par, std::string& regex)
+static void convert_escape_char(const char *pattern, size_t len, size_t& loc, size_t& pos, convert_flag_type flags, const char *signature, const std::map<size_t,std::string>& mod, const char *par, std::string& regex)
 {
   int c = pattern[pos];
   if (std::strchr(regex_unescapes, c) != NULL)
@@ -977,7 +1038,7 @@ static void convert_escape_char(const char *pattern, size_t& loc, size_t& pos, c
     int esc = hex_or_octal_escape(signature);
     if (is_modified(mod, 'u'))
     {
-      if (!supports_escape(signature, invert ? 'P' : 'p'))
+      if (!supports_escape(signature, 'p'))
         translated = unicode_class(name, esc, flags, par);
     }
     else if (!supports_escape(signature, c))
@@ -1075,6 +1136,13 @@ static void convert_escape_char(const char *pattern, size_t& loc, size_t& pos, c
         loc = pos + 1;
       }
     }
+    else if ((c == 'g' || c == 'k') && pos + 2 < len && pattern[pos + 1] == '{')
+    {
+      while (pos + 1 < len && pattern[pos + 1] != '\0' && pattern[++pos] != '}')
+        continue;
+      if (pos >= len)
+        throw regex_error(regex_error::mismatched_braces, pattern, pos);
+    }
   }
 }
 
@@ -1120,17 +1188,25 @@ static void convert_escape(const char *pattern, size_t len, size_t& loc, size_t&
   else if (c == 'N')
   {
     if (is_modified(mod, 'u'))
-      regex.append(&pattern[loc], pos - loc - 1).append(par).append("[^\\n][\\x80-\\xbf]*)");
+    {
+      if (!supports_escape(signature, 'p'))
+      {
+        regex.append(&pattern[loc], pos - loc - 1).append(par).append("[^\\n][\\x80-\\xbf]*)");
+        loc = pos + 1;
+      }
+    }
     else if (!supports_escape(signature, c))
+    {
       regex.append(&pattern[loc], pos - loc - 1).append("[^\n]");
-    loc = pos + 1;
+      loc = pos + 1;
+    }
   }
   else if (c >= '0' && c <= '7' && pattern[pos + 1] >= '0' && pattern[pos + 1] <= '7')
   {
     size_t k = pos;
     size_t n = k + 3 + (pattern[k] == '0');
     int wc = 0;
-    while (k < n && (c = pattern[k]) >= '0' && c <= '7')
+    while (k < n && k < len && (c = pattern[k]) >= '0' && c <= '7')
     {
       wc = 8 * wc + c - '0';
       ++k;
@@ -1156,10 +1232,10 @@ static void convert_escape(const char *pattern, size_t len, size_t& loc, size_t&
     pos = k - 1;
     loc = pos + 1;
   }
-  else if (c == 'u' || c == 'x')
+  else if (c == 'o' || c == 'u' || c == 'x')
   {
     size_t k = pos;
-    int wc = convert_hex(pattern, len, k, flags);
+    int wc = (c == 'o' ? convert_oct(pattern, len, k) : convert_hex(pattern, len, k, flags));
     if (wc >= 0)
     {
       if (c == 'u' && wc >= 0xD800 && wc < 0xE000)
@@ -1218,7 +1294,7 @@ static void convert_escape(const char *pattern, size_t len, size_t& loc, size_t&
     }
     else
     {
-      convert_escape_char(pattern, loc, pos, flags, signature, mod, par, regex);
+      convert_escape_char(pattern, len, loc, pos, flags, signature, mod, par, regex);
     }
   }
   else if (c == 'p' || c == 'P')
@@ -1236,6 +1312,8 @@ static void convert_escape(const char *pattern, size_t len, size_t& loc, size_t&
       k = j;
       while (k < len && pattern[k] != '}')
         ++k;
+      if (k >= len)
+        throw regex_error(regex_error::mismatched_braces, pattern, pos);
       name.append(pattern, j, k - j);
     }
     else
@@ -1246,46 +1324,65 @@ static void convert_escape(const char *pattern, size_t len, size_t& loc, size_t&
     }
     std::string translated;
     int esc = hex_or_octal_escape(signature);
-    if (is_modified(mod, 'u'))
+    if (supports_escape(signature, 'p'))
     {
-      translated = unicode_class(name.c_str(), esc, flags, par);
-      if (translated.empty())
+      translated = posix_class(name.c_str(), esc);
+      if (!translated.empty())
       {
-        translated = posix_class(name.c_str(), esc);
-        if (translated.empty())
-          throw regex_error(regex_error::invalid_class, pattern, pos);
-      }
-      else if (supports_escape(signature, c))
-      {
-        translated.clear(); // assume regex lib supports this Unicode character class
+        regex.append(&pattern[loc], pos - loc - 2).append("[").append(translated.substr(1, translated.size() - 2)).append("]");
+        loc = k + 1;
       }
     }
     else
     {
-      translated = posix_class(name.c_str(), esc);
-      if (translated.empty() && !supports_escape(signature, c))
-        throw regex_error(regex_error::invalid_class, pattern, pos);
-    }
-    if (!translated.empty())
-    {
-      regex.append(&pattern[loc], pos - loc - 2).append(translated);
-      loc = k + 1;
+      if (is_modified(mod, 'u'))
+      {
+        translated = unicode_class(name.c_str(), esc, flags, par);
+        if (translated.empty())
+        {
+          translated = posix_class(name.c_str(), esc);
+          if (translated.empty())
+            throw regex_error(regex_error::invalid_class, pattern, pos);
+        }
+      }
+      else
+      {
+        translated = posix_class(name.c_str(), esc);
+        if (translated.empty() && !supports_escape(signature, c))
+          throw regex_error(regex_error::invalid_class, pattern, pos);
+      }
+      if (!translated.empty())
+      {
+        regex.append(&pattern[loc], pos - loc - 2).append(translated);
+        loc = k + 1;
+      }
     }
     pos = k;
   }
   else if (c == 's' && (flags & convert_flag::notnewline))
   {
     // \s is the same as \p{Space} but without newline \n
-    ORanges<int> ranges;
-    insert_escape_class(pattern, pos, mod, ranges);
-    ranges.erase('\n');
-    regex.append(&pattern[loc], pos - loc - 1);
-    regex.append(convert_ranges(pattern, pos, ranges, mod, flags, signature, par));
-    loc = ++pos;
+    if (supports_escape(signature, 'p'))
+    {
+      if (is_modified(mod, 'u'))
+        regex.append(&pattern[loc], pos - loc - 1).append("[\\t\\x0b-\\r\\x85\\p{Z}]");
+      else
+        regex.append(&pattern[loc], pos - loc - 1).append("[\\h\\x0b-\\r\\x85\\xa0]");
+      loc = pos + 1;
+    }
+    else
+    {
+      ORanges<int> ranges;
+      insert_escape_class(pattern, pos, mod, ranges);
+      ranges.erase('\n');
+      regex.append(&pattern[loc], pos - loc - 1);
+      regex.append(convert_ranges(pattern, pos, ranges, mod, flags, signature, par));
+      loc = pos + 1;
+    }
   }
   else
   {
-    convert_escape_char(pattern, loc, pos, flags, signature, mod, par, regex);
+    convert_escape_char(pattern, len, loc, pos, flags, signature, mod, par, regex);
   }
 }
 
@@ -1339,8 +1436,6 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
         if (supports_modifier(signature, pattern[k]))
         {
           mods.push_back(pattern[k]);
-          if (pattern[k] == 'i' && (flags & convert_flag::unicode))
-            enable_modifier(pattern[k], pattern, k, mod, lev);
         }
         else if (pattern[k] == 'm')
         {
@@ -1530,12 +1625,18 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
                 }
                 --lev;
               }
+              else if (pattern[pos] == '(')
+              {
+                if (!supports_modifier(signature, '('))
+                  throw regex_error(regex_error::invalid_syntax, pattern, pos);
+                --pos;
+              }
               else
               {
                 std::string mods, unmods;
                 size_t k = pos;
                 bool invert = false;
-                while (k < len && (pattern[k] == '-' || std::isalpha(pattern[k])))
+                while (k < len && (pattern[k] == '-' || std::isalnum(pattern[k])))
                 {
                   if (pattern[k] == '-')
                   {
@@ -1559,7 +1660,7 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
                 }
                 if (k >= len)
                   throw regex_error(regex_error::mismatched_parens, pattern, pos);
-                if (pattern[k] == ':' || pattern[k] == '>' || pattern[k] == ')')
+                if (pattern[k] == ':' || pattern[k] == ')')
                 {
                   regex.append(&pattern[loc], pos - loc - 2);
                   if (pattern[k] == ')')
@@ -1578,7 +1679,7 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
                     mod[lev].clear();
                     --lev;
                   }
-                  else if (pattern[k] == ':')
+                  else
                   {
                     // (?imsx:...)
                     if (can)
@@ -1606,6 +1707,14 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
               }
             }
           }
+          else if (pos + 1 < len && pattern[pos + 1] == '*' && supports_modifier(signature, '*'))
+          {
+            pos += 2;
+            while (pos < len && pattern[pos] != ')')
+              ++pos;
+            if (pos < len)
+              --lev;
+          }
           else
           {
             beg = true;
@@ -1828,7 +1937,7 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
             {
               regex.append(par).append(subregex).push_back(')');
             }
-            loc = pos + (pattern[pos] == '\\') + 1;
+            loc = pos + (pos < len && pattern[pos] == '\\') + 1;
             anc = false;
             beg = false;
           }
@@ -1844,7 +1953,7 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
             char *s;
             size_t n = static_cast<size_t>(std::strtoul(&pattern[pos], &s, 10));
             pos = s - pattern;
-            if (pattern[pos] == ',')
+            if (pos + 1 < len && pattern[pos] == ',')
             {
               ++pos;
               if (pattern[pos] != '\\' && pattern[pos] != '}')
@@ -1913,16 +2022,23 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
         }
         break;
       case '.':
-        if (is_modified(mod, 'u') && pos + 1 < len && pattern[pos + 1] != '*' && pattern[pos + 1] != '+')
+        if (is_modified(mod, 'u') && (pos + 1 >= len || (pattern[pos + 1] != '*' && pattern[pos + 1] != '+')))
         {
           // unicode: translate . to match any UTF-8 so . matches anything (also beyond U+10FFFF)
           if (is_modified(mod, 's'))
+          {
             regex.append(&pattern[loc], pos - loc).append(par).append("[\\x00-\\xff][\\x80-\\xbf]*)");
-          else if (supports_modifier(signature, 's') || supports_escape(signature, '.'))
-            regex.append(&pattern[loc], pos - loc).append(par).append(".[\\x80-\\xbf]*)");
-          else
-            regex.append(&pattern[loc], pos - loc).append(par).append("[^\\n][\\x80-\\xbf]*)");
-          loc = pos + 1;
+            loc = pos + 1;
+          }
+          else if (!supports_escape(signature, 'p'))
+          {
+            // \p is not supported: this indicates that . is non-Unicode
+            if (supports_modifier(signature, 's') || supports_escape(signature, '.'))
+              regex.append(&pattern[loc], pos - loc).append(par).append(".[\\x80-\\xbf]*)");
+            else
+              regex.append(&pattern[loc], pos - loc).append(par).append("[^\\n][\\x80-\\xbf]*)");
+            loc = pos + 1;
+          }
         }
         else if (is_modified(mod, 's'))
         {
@@ -1996,14 +2112,22 @@ std::string convert(const char *pattern, const char *signature, convert_flag_typ
             loc = pos + 1;
           }
         }
-        else if ((c & 0xC0) == 0xC0 && is_modified(mod, 'u'))
+        else if ((c & 0xC0) == 0xC0 && is_modified(mod, 'u') && !supports_escape(signature, 'p'))
         {
           // unicode: group UTF-8 sequence
-          regex.append(&pattern[loc], pos - loc).append(par);
-          while (((c = pattern[++pos]) & 0xC0) == 0x80)
-            continue;
-          regex.append(pattern + loc, pos - loc).push_back(')');
+          regex.append(&pattern[loc], pos - loc);
           loc = pos;
+          while (pos + 1 < len && ((c = pattern[++pos]) & 0xC0) == 0x80)
+            continue;
+          if (pos < len &&
+              (pattern[pos] == '*' ||
+               ((flags & convert_flag::basic) && pos + 1 < len ?
+                (pattern[pos] == '\\' && (pattern[pos + 1] == '?' || pattern[pos + 1] == '+' || pattern[pos + 1] == '{')) :
+                (pattern[pos] == '?' || pattern[pos] == '+' || pattern[pos] == '{'))))
+          {
+            regex.append(par).append(&pattern[loc], pos - loc).push_back(')');
+            loc = pos;
+          }
           --pos;
         }
         anc = false;
