@@ -52,7 +52,8 @@ searches.
 .PP
 A query interface is opened with \fB-Q\fR (\fB--query\fR) to interactively
 specify search patterns and view search results.  Note that a \fIPATTERN\fR
-argument should not be specified in this case.
+argument cannot be specified in this case.  To specify one or more patterns
+with \fB-Q\fR, use \fB-e PATTERN\fR.
 .PP
 For help, \fB--help \fIWHAT\fR displays help on options related to \fIWHAT\fR.
 .PP
@@ -220,10 +221,6 @@ by a previous glob pattern specified in the files associated with
 .IP \fBGREP_PATH\fR
 May be used to specify a file path to pattern files.  The file path is used by
 option \fB-f\fR to open a pattern file, when the pattern file does not exist.
-.IP \fBGREP_EDITOR\fR
-May be used to specify an editor command to invoke with CTRL-Y while using the
-query UI with option \fB-Q\fR.  When undefined, the command defined by
-\fBEDITOR\fR is invoked.
 .IP \fBGREP_COLOR\fR
 May be used to specify ANSI SGR parameters to highlight matches when option
 \fB--color\fR is used, e.g. 1;35;40 shows pattern matches in bold magenta text
@@ -259,7 +256,7 @@ default.
 SGR substring for matching text in a context line.  The substring \fBmt=\fR by
 default.
 .IP \fBfn=\fR
-SGR substring for file names.
+SGR substring for filenames.
 .IP \fBln=\fR
 SGR substring for line numbers.
 .IP \fBcn=\fR
@@ -268,6 +265,12 @@ SGR substring for column numbers.
 SGR substring for byte offsets.
 .IP \fBse=\fR
 SGR substring for separators.
+.IP \fBrv\fR
+a Boolean parameter, switches \fBsl=\fR and \fBcx=\fR with option \fB-v\fR.
+.IP \fBhl\fR
+a Boolean parameter, enables filename hyperlinks (\fB\\33]8;;link\fR).
+.IP \fBne\fR
+a Boolean parameter, disables ``erase in line'' \fB\\33[K\fR.
 .SH FORMAT
 Option \fB--format\fR=\fIFORMAT\fR specifies an output format for file matches.
 Fields may be used in \fIFORMAT\fR, which expand into the following values:
