@@ -226,16 +226,29 @@ void Query::draw()
   {
     if (select_ == -1)
     {
+      start_ = 0;
       Screen::home();
 
-      start_ = static_cast<int>(prompt_.size());
+      if (row_ > 0)
+      {
+        char down[16];
+
+        snprintf(down, sizeof(down), "%3d ", row_);
+
+        if (!Screen::mono)
+          Screen::put(PROMPT);
+        Screen::put(down);
+
+        start_ = strlen(down);
+      }
 
       if (!dirs_.empty())
       {
         int width = Screen::mbstring_width(dirs_.c_str());
         int offset = 0;
-        if (width + 2 > Screen::cols/2)
-          offset = width + 2 - Screen::cols/2;
+        int split = Screen::cols/2 - start_;
+        if (width + 2 > split)
+          offset = width + 2 - split;
 
         Screen::normal();
         if (offset > 0)
@@ -261,6 +274,8 @@ void Query::draw()
 
       Screen::normal();
 
+      start_ += static_cast<int>(prompt_.size());
+
       int pos;
       if (len_ - col_ < shift_)
         pos = Screen::cols - start_ - (len_ - col_) - 1;
@@ -401,7 +416,7 @@ void Query::draw()
 }
 
 // view one row of text by drawing it on screen
-void Query::view(int row)
+void Query::disp(int row)
 {
   Screen::normal();
   if (selected_[row])
@@ -440,14 +455,14 @@ void Query::redraw()
     Screen::put(13, 0, "\033[7m^K\033[m delete after cursor");
     Screen::put(14, 0, "\033[7m^L\033[m refresh screen");
     Screen::put(15, 0, "\033[7m^Q\033[m quick exit and output");
-    Screen::put(16, 0, "\033[7m^R\033[m or \033[7mF4\033[m jump to bookmark");
+    Screen::put(16, 0, "\033[7m^R\033[m or \033[7mF4\033[m restore bookmark");
     Screen::put(17, 0, "\033[7m^S\033[m scroll to next file/dir");
     Screen::put(18, 0, "\033[7m^T\033[m toggle colors on/off");
     Screen::put(19, 0, "\033[7m^U\033[m delete before cursor");
     Screen::put(20, 0, "\033[7m^V\033[m verbatim character");
     Screen::put(21, 0, "\033[7m^W\033[m scroll back one file/dir");
     Screen::put(22, 0, "\033[7m^X\033[m or \033[7mF3\033[m set bookmark");
-    Screen::put(23, 0, "\033[7m^Y\033[m or \033[7mF2\033[m edit file");
+    Screen::put(23, 0, "\033[7m^Y\033[m or \033[7mF2\033[m view or edit file");
     Screen::put(24, 0, "\033[7m^Z\033[m or \033[7mF1\033[m help");
     Screen::put(25, 0, "\033[7m^^\033[m chdir to starting dir");
     Screen::put(26, 0, "\033[7m^\\\033[m terminate process");
@@ -509,7 +524,7 @@ void Query::redraw()
     if (end > row_ + Screen::rows - 1)
       end = row_ + Screen::rows - 1;
     for (int i = row_; i < end; ++i)
-      view(i);
+      disp(i);
     if (!message_)
       draw();
   }
@@ -584,9 +599,9 @@ void Query::move(int col)
 void Query::insert(const char *text, size_t size)
 {
   char *end = line_end();
-  if (end + size >= line_ + QUERY_MAX_LEN)
+  if (end + size >= line_ + sizeof(Line))
   {
-    size = line_ + QUERY_MAX_LEN - end - 1;
+    size = line_ + sizeof(Line) - end - 1;
     Screen::alert();
   }
   if (size > 0)
@@ -754,8 +769,8 @@ void Query::query_ui()
     flag_regexp.clear();
 
     size_t num = pattern.size();
-    if (num >= QUERY_MAX_LEN)
-      num = QUERY_MAX_LEN - 1;
+    if (num >= sizeof(Line))
+      num = sizeof(Line) - 1;
 
     pattern.copy(line_, num);
     line_[num] = '\0';
@@ -880,7 +895,7 @@ void Query::query_ui()
             {
               globbing_ = false;
 
-              memcpy(line_, temp_, QUERY_MAX_LEN);
+              memcpy(line_, temp_, sizeof(Line));
               len_ = line_len();
               move(len_);
 
@@ -920,7 +935,7 @@ void Query::query_ui()
             else
             {
               selected_[select_] = !selected_[select_];
-              view(select_);
+              disp(select_);
               down();
             }
           }
@@ -1036,7 +1051,7 @@ void Query::query_ui()
             {
               up();
               selected_[select_] = !selected_[select_];
-              view(select_);
+              disp(select_);
             }
           }
           else if (mode_ == Mode::EDIT)
@@ -1063,7 +1078,7 @@ void Query::query_ui()
           {
             up();
             selected_[select_] = !selected_[select_];
-            view(select_);
+            disp(select_);
           }
           break;
 
@@ -1156,7 +1171,22 @@ void Query::query_ui()
 
         case VKey::CTRL_R: // CTRL-R: jump to bookmark
         case VKey::FN(4):
-          jump(mark_);
+          if (mark_.row >= 0)
+          {
+            int row;
+            if (mark_.restore(line_, col_, row, flags_))
+            {
+              globbing_ = false;
+              set_prompt();
+              len_ = line_len();
+              search();
+            }
+            jump(row);
+          }
+          else
+          {
+            Screen::alert();
+          }
           break;
 
         case VKey::CTRL_Q: // CTRL-Q: immediately quit and output lines
@@ -1195,14 +1225,17 @@ void Query::query_ui()
           back();
           break;
 
-        case VKey::CTRL_X: // CTRL-X: set bookmark
+        case VKey::CTRL_X: // CTRL-X: set bookmark and save state
         case VKey::FN(3):
-          mark_ = select_ >= 0 ? select_ : row_;
+          mark_.save(line_, col_, (select_ >= 0 ? select_ : row_), flags_);
           break;
 
         case VKey::CTRL_Y: // CTRL-Y: edit file
         case VKey::FN(2):
-          edit();
+          if (select_ == -1)
+            view();
+          else
+            Screen::alert();
           break;
 
         case VKey::CTRL_Z: // CTRL-Z: help
@@ -1377,7 +1410,7 @@ bool Query::update()
       begin = row_;
 
     for (int i = begin; i < end; ++i)
-      view(i);
+      disp(i);
   }
 
   // display final status line if in view
@@ -1674,8 +1707,8 @@ void Query::load_line()
     if (static_cast<size_t>(select_) < view_.size())
     {
       size_t size = view_[select_].size();
-      if (size >= QUERY_MAX_LEN)
-        size = QUERY_MAX_LEN - 1;
+      if (size >= sizeof(Line))
+        size = sizeof(Line) - 1;
       view_[select_].copy(line_, size);
       line_[size] = '\0';
       len_ = line_len();
@@ -1715,7 +1748,7 @@ void Query::up()
   }
   if (row_ > 0)
   {
-    view(row_ - 1);
+    disp(row_ - 1);
     --row_;
     Screen::pan_down();
     draw();
@@ -1740,7 +1773,7 @@ void Query::down()
     Screen::normal();
     Screen::pan_up();
     if (row_ + Screen::rows - 2 < rows_)
-      view(row_ + Screen::rows - 2);
+      disp(row_ + Screen::rows - 2);
     draw();
   }
 }
@@ -1762,7 +1795,7 @@ void Query::pgup(bool half_page)
   }
   if (row_ > 0)
   {
-    view(row_ - 1);
+    disp(row_ - 1);
     int oldrow = row_;
     if (half_page)
       row_ -= Screen::rows / 2;
@@ -1772,7 +1805,7 @@ void Query::pgup(bool half_page)
       row_ = 0;
     Screen::pan_down(oldrow - row_);
     for (int i = row_; i < oldrow - 1; ++i)
-      view(i);
+      disp(i);
     draw();
   }
 }
@@ -1812,7 +1845,7 @@ void Query::pgdn(bool half_page)
       Screen::pan_up(diff);
       for (int i = row_ + Screen::rows - diff - 1; i < row_ + Screen::rows - 1; ++i)
         if (i < rows_)
-          view(i);
+          disp(i);
       draw();
     }
   }
@@ -1971,9 +2004,6 @@ void Query::next()
 // jump to the specified row
 void Query::jump(int row)
 {
-  if (rows_ <= 0)
-    return;
-
   if (row < 0)
     row = 0;
 
@@ -1997,6 +2027,7 @@ void Query::jump(int row)
         while (row_ < row && row_ + 1 < rows_)
           ++row_;
 
+        // exit if at the desired row or if the desired row is beyond the end of the search results
         if (row_ == row || (eof_ && buflen_ == 0))
           break;
 
@@ -2038,6 +2069,7 @@ void Query::jump(int row)
         while (select_ < row && select_ + 1 < rows_)
           ++select_;
 
+        // exit if at the desired row or if the desired row is beyond the end of the search results
         if (select_ == row || (eof_ && buflen_ == 0))
           break;
 
@@ -2063,8 +2095,8 @@ void Query::jump(int row)
   redraw();
 }
 
-// edit the file located under the cursor or just above in the screen
-void Query::edit()
+// view or edit the file located under the cursor or just above in the screen (when not in selection mode)
+void Query::view()
 {
   if (row_ >= rows_ || flag_text || flag_format != NULL)
   {
@@ -2073,12 +2105,26 @@ void Query::edit()
     return;
   }
 
-  const char *editor = getenv("GREP_EDIT");
+  if (flag_stdin)
+  {
+    message("cannot view or edit standard input");
+    return;
+  }
 
-  if (editor == NULL)
-    editor = getenv("EDITOR");
+  const char *pager = flag_view;
 
-  if (editor == NULL)
+  if (pager != NULL && *pager == '\0')
+  {
+    pager = getenv("PAGER");
+
+    if (pager == NULL)
+      pager = getenv("EDITOR");
+
+    if (pager == NULL)
+      pager = DEFAULT_VIEW_COMMAND;
+  }
+
+  if (pager == NULL || *pager == '\0')
   {
     Screen::alert();
 
@@ -2087,8 +2133,9 @@ void Query::edit()
 
   std::string filename;
   bool found = false;
+  int row;
 
-  for (int i = select_ >= 0 ? select_ : row_; i >= 0 && !(found = is_filename(view_[i], filename)); --i)
+  for (row = row_; row >= 0 && !(found = is_filename(view_[row], filename)); --row)
     continue;
 
   if (!found && arg_files.size() == 1)
@@ -2100,36 +2147,71 @@ void Query::edit()
   if (found)
   {
 #ifdef OS_WIN
-    DWORD attr = GetFileAttributesA(filename.c_str());
-    found = attr != INVALID_FILE_ATTRIBUTES && !(attr & FILE_ATTRIBUTE_DIRECTORY) && !(attr & FILE_ATTRIBUTE_SYSTEM);
+    _WIN32_FILE_ATTRIBUTE_DATA attr_before;
+    found = GetFileAttributesExW(utf8_decode(filename).c_str(), GetFileExInfoStandard, &attr_before);
 #else
     struct stat buf;
     found = stat(filename.c_str(), &buf) == 0 && S_ISREG(buf.st_mode);
+#if defined(HAVE_STAT_ST_ATIM) && defined(HAVE_STAT_ST_MTIM) && defined(HAVE_STAT_ST_CTIM)
+    uint64_t mtime = static_cast<uint64_t>(buf.st_mtim.tv_sec) * 1000000 + buf.st_mtim.tv_nsec / 1000;
+#elif defined(HAVE_STAT_ST_ATIMESPEC) && defined(HAVE_STAT_ST_MTIMESPEC) && defined(HAVE_STAT_ST_CTIMESPEC)
+    uint64_t mtime = static_cast<uint64_t>(buf.st_mtimespec.tv_sec) * 1000000 + buf.st_mtimespec.tv_nsec / 1000;
+#else
+    uint64_t mtime = static_cast<uint64_t>(buf.st_mtime);
+#endif
 #endif
-  }
 
-  if (found)
-  {
-    std::string command;
-    command.assign(editor).append(" \"").append(filename).append("\"");
-
-    Screen::put(0, 0, command.c_str());
-
-    if (system(command.c_str()) == 0)
+    if (found)
     {
-      mark_ = select_ >= 0 ? select_ : row_;
+      std::string command(pager);
+
+      command.append(" \"").append(filename).append("\"");
+
       Screen::clear();
-      search();
-    }
-    else
-    {
-      Screen::alert();
+
+      if (system(command.c_str()) == 0)
+      {
+        Screen::clear();
+
+        bool changed;
+
+#ifdef OS_WIN
+        _WIN32_FILE_ATTRIBUTE_DATA attr_after;
+        changed = GetFileAttributesExW(utf8_decode(filename).c_str(), GetFileExInfoStandard, &attr_after) == 0 ||
+          attr_before.ftLastWriteTime.dwLowDateTime != attr_after.ftLastWriteTime.dwLowDateTime ||
+          attr_before.ftLastWriteTime.dwHighDateTime != attr_after.ftLastWriteTime.dwHighDateTime;
+#else
+        stat(filename.c_str(), &buf);
+#if defined(HAVE_STAT_ST_ATIM) && defined(HAVE_STAT_ST_MTIM) && defined(HAVE_STAT_ST_CTIM)
+        changed = (mtime != static_cast<uint64_t>(buf.st_mtim.tv_sec) * 1000000 + buf.st_mtim.tv_nsec / 1000);
+#elif defined(HAVE_STAT_ST_ATIMESPEC) && defined(HAVE_STAT_ST_MTIMESPEC) && defined(HAVE_STAT_ST_CTIMESPEC)
+        changed = (mtime != static_cast<uint64_t>(buf.st_mtimespec.tv_sec) * 1000000 + buf.st_mtimespec.tv_nsec / 1000);
+#else
+        changed = (mtime != static_cast<uint64_t>(buf.st_mtime));
+#endif
+#endif
+
+        if (changed)
+        {
+          // file is changed, update the search results
+          search();
+          jump(row);
+        }
+        else
+        {
+          redraw();
+        }
+      }
+      else
+      {
+        Screen::put(0, 0, command.c_str());
+        Screen::alert();
+      }
     }
   }
-  else
-  {
+
+  if (!found)
     message(std::string("cannot edit file ").append(filename));
-  }
 }
 
 // chdir one level down into the directory of the file located under the cursor or just above the screen
@@ -2157,27 +2239,34 @@ void Query::select()
 
   if (found)
   {
+    if (globbing_)
+    {
+      globbing_ = false;
+
+      memcpy(line_, temp_, sizeof(Line));
+      len_ = line_len();
+      move(len_);
+
+      set_prompt();
+    }
+
+    history_.emplace();
+    history_.top().save(line_, col_, row_, flags_, mark_);
+
+    mark_.unset();
+
     size_t n = pathname.find(PATHSEPCHR);
     size_t b = pathname.find('{'); // do not cd into archives
     if (n != std::string::npos && (b == std::string::npos || n < b))
     {
       pathname.resize(n);
 
-#ifdef OS_WIN
-      if (_chdir(pathname.c_str()) < 0)
-      {
-        message("cannot chdir: operation denied");
-
-        return;
-      }
-#else
       if (chdir(pathname.c_str()) < 0)
       {
         message("cannot chdir: operation denied");
 
         return;
       }
-#endif
 
       dirs_.append(pathname).append(PATHSEPSTR);
 
@@ -2230,11 +2319,7 @@ void Query::deselect()
 
     if (dirs_.empty())
     {
-#ifdef OS_WIN
-      char *cwd = _getcwd(NULL, 0);
-#else
-      char *cwd = getcwd(NULL, 0);
-#endif
+      char *cwd = getcwd0();
       if (cwd != NULL)
       {
         size_t n = strlen(cwd);
@@ -2246,13 +2331,8 @@ void Query::deselect()
       }
     }
 
-#ifdef OS_WIN
-    if (_chdir("..") < 0)
-      return;
-#else
     if (chdir("..") < 0)
       return;
-#endif
 
     if (dirs_.empty())
     {
@@ -2280,8 +2360,6 @@ void Query::deselect()
         dirs_.resize(n + 1);
       }
     }
-
-    search();
   }
   else
   {
@@ -2293,7 +2371,23 @@ void Query::deselect()
       dirs_.clear();
 
     deselect_file_ = true;
+  }
 
+  mark_.unset();
+
+  if (!history_.empty())
+  {
+    int row;
+    history_.top().restore(line_, col_, row, flags_, mark_);
+    history_.pop();
+    globbing_ = false;
+    set_prompt();
+    len_ = line_len();
+    search();
+    jump(row);
+  }
+  else
+  {
     search();
   }
 }
@@ -2303,20 +2397,8 @@ void Query::unselect()
 {
   if (!wdir_.empty())
   {
-#ifdef OS_WIN
-    if (_chdir(wdir_.c_str()) < 0)
-      return;
-#else
     if (chdir(wdir_.c_str()) < 0)
       return;
-#endif
-
-    dirs_.clear();
-    wdir_.clear();
-
-    deselect_file_ = true;
-
-    search();
   }
   else if (!dirs_.empty())
   {
@@ -2335,7 +2417,7 @@ void Query::unselect()
       while (true)
       {
 #ifdef OS_WIN
-        if ((dirs_.size() == 3 && dirs_.at(1) == ':' && dirs_.at(2) == PATHSEPCHR) || _chdir("..") < 0)
+        if ((dirs_.size() == 3 && dirs_.at(1) == ':' && dirs_.at(2) == PATHSEPCHR) || chdir("..") < 0)
           break;
 #else
         if (dirs_ == PATHSEPSTR || chdir("..") < 0)
@@ -2351,12 +2433,30 @@ void Query::unselect()
 
         dirs_.resize(n + 1);
       }
-
-      dirs_.clear();
     }
+  }
 
-    deselect_file_ = true;
+  dirs_.clear();
+  wdir_.clear();
+  deselect_file_ = true;
 
+  mark_.unset();
+
+  if (!history_.empty())
+  {
+    while (history_.size() > 1)
+      history_.pop();
+    int row;
+    history_.top().restore(line_, col_, row, flags_, mark_);
+    history_.pop();
+    globbing_ = false;
+    set_prompt();
+    len_ = line_len();
+    search();
+    jump(row);
+  }
+  else
+  {
     search();
   }
 }
@@ -2365,7 +2465,7 @@ void Query::unselect()
 void Query::message(const std::string& message)
 {
   if (!Screen::mono)
-   Screen::put(PROMPT);
+    Screen::put(PROMPT);
   Screen::put(0, 0, "-> ");
   if (!Screen::mono)
     Screen::normal();
@@ -2773,10 +2873,10 @@ void Query::meta(int key)
           {
             globbing_ = true;
 
-            memcpy(temp_, line_, QUERY_MAX_LEN);
+            memcpy(temp_, line_, sizeof(Line));
             size_t num = globs_.size();
-            if (num >= QUERY_MAX_LEN)
-              num = QUERY_MAX_LEN - 1;
+            if (num >= sizeof(Line))
+              num = sizeof(Line) - 1;
             memcpy(line_, globs_.c_str(), num);
             line_[num] = '\0';
 
@@ -2789,7 +2889,7 @@ void Query::meta(int key)
           {
             globbing_ = false;
 
-            memcpy(line_, temp_, QUERY_MAX_LEN);
+            memcpy(line_, temp_, sizeof(Line));
             len_ = line_len();
             move(len_);
 
@@ -3349,7 +3449,7 @@ ssize_t Query::stdin_sender(int fd)
   return nwritten;
 }
 
-// true if line starts with a valid filename/filepath identified by three \0 markers
+// true if line starts with a valid filename/filepath identified by three \0 markers and differs from the given filename, then assigns filename
 bool Query::is_filename(const std::string& line, std::string& filename, bool compare_dir)
 {
   size_t start = 0;
@@ -3425,55 +3525,56 @@ bool Query::is_filename(const std::string& line, std::string& filename, bool com
   return true;
 }
 
-Query::Mode              Query::mode_                = Query::Mode::QUERY;
-bool                     Query::updated_             = false;
-bool                     Query::message_             = false;
-char                     Query::line_[QUERY_MAX_LEN] = { '\0' };
-char                     Query::temp_[QUERY_MAX_LEN] = { '\0' };
-std::string              Query::prompt_;
-int                      Query::start_               = 0;
-int                      Query::col_                 = 0;
-int                      Query::len_                 = 0;
-int                      Query::offset_              = 0;
-int                      Query::shift_               = 8;
-std::atomic_int          Query::error_;
-std::string              Query::what_;
-int                      Query::row_                 = 0;
-int                      Query::rows_                = 0;
-int                      Query::mark_                = 0;
-int                      Query::select_              = -1;
-bool                     Query::select_all_          = false;
-bool                     Query::globbing_            = false;
-std::string              Query::globs_;
-std::string              Query::dirs_;
-std::string              Query::wdir_;
-bool                     Query::deselect_file_;
-std::string              Query::selected_file_;
-int                      Query::skip_                = 0;
-std::vector<std::string> Query::view_;
-std::list<std::string>   Query::saved_;
-std::vector<bool>        Query::selected_;
-bool                     Query::eof_                 = true;
-bool                     Query::append_              = false;
-size_t                   Query::buflen_              = 0;
-char                     Query::buffer_[QUERY_BUFFER_SIZE];
-int                      Query::search_pipe_[2];
-std::thread              Query::search_thread_;
-std::string              Query::stdin_buffer_;
-int                      Query::stdin_pipe_[2];
-std::thread              Query::stdin_thread_;
-char                     Query::searching_[16]       = "Searching...";
-int                      Query::dots_                = 3;
-size_t                   Query::context_             = 2;
-size_t                   Query::fuzzy_               = 1;
-bool                     Query::dotall_              = false;
+Query::Mode                Query::mode_                = Query::Mode::QUERY;
+bool                       Query::updated_             = false;
+bool                       Query::message_             = false;
+Query::Line                Query::line_                = { '\0' };
+Query::Line                Query::temp_                = { '\0' };
+std::string                Query::prompt_;
+int                        Query::start_               = 0;
+int                        Query::col_                 = 0;
+int                        Query::len_                 = 0;
+int                        Query::offset_              = 0;
+int                        Query::shift_               = 8;
+std::atomic_int            Query::error_;
+std::string                Query::what_;
+int                        Query::row_                 = 0;
+int                        Query::rows_                = 0;
+Query::State               Query::mark_;
+int                        Query::select_              = -1;
+bool                       Query::select_all_          = false;
+bool                       Query::globbing_            = false;
+std::string                Query::globs_;
+std::string                Query::dirs_;
+std::string                Query::wdir_;
+bool                       Query::deselect_file_;
+std::string                Query::selected_file_;
+std::stack<Query::History> Query::history_;
+int                        Query::skip_                = 0;
+std::vector<std::string>   Query::view_;
+std::list<std::string>     Query::saved_;
+std::vector<bool>          Query::selected_;
+bool                       Query::eof_                 = true;
+bool                       Query::append_              = false;
+size_t                     Query::buflen_              = 0;
+char                       Query::buffer_[QUERY_BUFFER_SIZE];
+int                        Query::search_pipe_[2];
+std::thread                Query::search_thread_;
+std::string                Query::stdin_buffer_;
+int                        Query::stdin_pipe_[2];
+std::thread                Query::stdin_thread_;
+char                       Query::searching_[16]       = "Searching...";
+int                        Query::dots_                = 3;
+size_t                     Query::context_             = 2;
+size_t                     Query::fuzzy_               = 1;
+bool                       Query::dotall_              = false;
 
 #ifdef OS_WIN
 
-HANDLE                   Query::hPipe_;
-OVERLAPPED               Query::overlapped_;
-bool                     Query::blocking_;
-bool                     Query::pending_;
+HANDLE                     Query::hPipe_;
+OVERLAPPED                 Query::overlapped_;
+bool                       Query::blocking_;
+bool                       Query::pending_;
 
 #endif
 
